#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : detect.py
# Contents: detect synchronous spiking activity and reduce results
# Author  : Christian Borgelt
# History : 2014.12.19 file created from cocoexp/detect.py
#           2014.12.22 reduction functions added (instead of patred.py)
#           2015.01.07 added code to create subdirectories
#-----------------------------------------------------------------------
from sys     import argv, path
from os.path import join
from os      import makedirs
from time    import strftime
from random  import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par  import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts    import genpsts
try:
    from ccnovl import coconad
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from coconad import coconad
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------

def patred (pats):
    pats.sort(key=lambda x: x[2], reverse=True)
    pall = pats                 # note the set of all patterns
    tmp  = []                   # initialize intermediary pattern list
    while pats:                 # while there are more patterns
        tmp.append(pats[0])     # transfer best pattern to output
        p,c,v = pats[0]         # remove all subset patterns
        pats  = [q for q in pats[1:] if not q[0] <= p]
    for p,c,v in tmp:           # traverse the intermediary patterns
        for q,x,y in pall:      # traverse all patterns
            if y < v: pats.append((p,c)); break
            if q < p: break     # remove all supersets that are worse
    return pats                 # return the reduced pattern list

#-----------------------------------------------------------------------

def redfn00 (pats):
    return pats

def redfn01 (pats):
    return patred([(p,c, len(p)   *c) for p,c in pats])

def redfn02 (pats):
    return patred([(p,c,(len(p)-1)*c) for p,c in pats])

def redfn03 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.2)) for p,c in pats])

def redfn04 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.4)) for p,c in pats])

def redfn05 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.6)) for p,c in pats])

def redfn06 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.8)) for p,c in pats])

def redfn07 (pats):
    return patred([(p,c,(len(p)-1)*(c+1.0)) for p,c in pats])

def redfn08 (pats):
    return patred([(p,c,(len(p)-1)*(c+1.2)) for p,c in pats])

def redfn09 (pats):
    return patred([(p,c,(len(p)-1)*(c+1.4)) for p,c in pats])

def redfn10 (pats):
    return patred([(p,c,(len(p)-1)*(c+1.6)) for p,c in pats])

def redfn11 (pats):
    return patred([(p,c,(len(p)-1)*(c+1.8)) for p,c in pats])

def redfn12 (pats):
    return patred([(p,c,(len(p)-1)*(c+2.0)) for p,c in pats])

def redfn13 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.02*len(p))) for p,c in pats])

def redfn14 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.04*len(p))) for p,c in pats])

def redfn15 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.06*len(p))) for p,c in pats])

def redfn16 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.08*len(p))) for p,c in pats])

def redfn17 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.10*len(p))) for p,c in pats])

def redfn18 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.12*len(p))) for p,c in pats])

def redfn19 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.14*len(p))) for p,c in pats])

def redfn20 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.15*len(p))) for p,c in pats])

def redfn21 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.16*len(p))) for p,c in pats])

def redfn22 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.18*len(p))) for p,c in pats])

def redfn23 (pats):
    return patred([(p,c,(len(p)-1)*(c+0.20*len(p))) for p,c in pats])

redfns = [redfn00, redfn01, redfn02, redfn03, redfn04, redfn05, redfn06,
          redfn07, redfn08, redfn09, redfn10, redfn11, redfn12, redfn13,
          redfn14, redfn15, redfn16, redfn17, redfn18, redfn19, redfn20,
          redfn21, redfn22, redfn23]

#-----------------------------------------------------------------------

def getredfn (id):
    if id < 0: return len(redfns)
    return redfns[id]

#-----------------------------------------------------------------------

def detect (ncnt=100, period=3, rate=20,         # neuron    parameters
            size=5, coin=5, jitter=0.003,        # assembly  parameters
            width=0.003, target='c', rfnids=[0], # detection parameters
            runs=1000, verbose=100, srng=1):     # other     parameters
    if jitter < 0: jitter = width
    srng += 1023*size           # init. seed and get data file name
    base = 'n%d_p%g_r%g_w%g_%c' % (ncnt,period,rate,width*10000,target)
    border = [0 for i in range(ncnt+1)]
    with open(join('data','pc','bd_'+base+'.txt'), 'r') as inp:
        for line in inp:        # collect the support values
            z,m = line.split(); border[int(z)] = float(m)
    for z in reversed(range(ncnt)):
        if border[z+1] > border[z]:  # ensure a monotone
            border[z] = border[z+1]  # detection border
    border[0:2] = [oo,oo]       # entirely rule out sizes 0 and 1
    base = 'n%d_p%g_r%g_j%g_w%g_%c' \
         % (ncnt,period,rate,jitter*10000,width*10000,target)
    ext  = '_%d_%d' % (size,coin)
    name = base +ext            # form the file name (parts)
    asm  = set(range(1,size+1)) # create the assembly to inject
    upcs = dict()               # and the pattern count dictionaries
    fpcs = dict()               # (unfiltered, filtered, reduced)
    rpcs = dict([(r,dict()) for r in rfnids])
    draw = xraw = 0                      # runs with raw detections
    dect = dict([(r,0) for r in rfnids]) # runs with detections
    xdct = dict([(r,0) for r in rfnids]) # runs with exact detections
    rexs = dict([(r,0) for r in rfnids]) # runs with excess  neurons
    rmis = dict([(r,0) for r in rfnids]) # runs with missing neurons
    rops = dict([(r,0) for r in rfnids]) # runs with other     patterns
    sups = dict([(r,0) for r in rfnids]) # number of superset  patterns
    subs = dict([(r,0) for r in rfnids]) # number of subset    patterns
    ovls = dict([(r,0) for r in rfnids]) # number of overlap   patterns
    urls = dict([(r,0) for r in rfnids]) # number of unrelated patterns
    for i in range(runs):       # repeat the experiment multiple times
        if verbose > 0 and i % verbose == 0:
            print(strftime('%Y.%m.%d %H:%M:%S detect ')+name+' '+str(i))
        seed(srng)              # seed random number generator
        psts = genpsts(ncnt, period, rate, jitter=jitter,
                       asms=[[asm,-coin]])
        pats = coconad(list(zip(range(1,ncnt+1), psts)),
                       target=target, width=width, supp=1.0,
                       zmin=2, zmax=-1, report='a', algo='i', mode='')
        fltd = []; d = e = 0    # generate and mine spike trains
        for p,c in pats:        # and traverse the found patterns
            z = len(p)          # get the size of the pattern
            p = frozenset(p)    # and the set of items/neurons
            if p >= asm:        # check for raw detections
                if z > size: e = 1     # superset detection
                else:        d = 1     # exact    detection
            s = (z,int(c*5))    # compute histogram bin
            if s in upcs: upcs[s] += 1 # count unfiltered patterns
            else:         upcs[s]  = 1 # by their signatures
            if c <= border[z]: continue
            if s in fpcs: fpcs[s] += 1 # count filtered patterns
            else:         fpcs[s]  = 1 # by their signatures
            fltd.append((p,c))  # collect non-excluded patterns
        draw += d | e           # count runs with raw detections
        xraw += d               # (superset and exact)
        for r in rfnids:        # traverse the reduction function ids
            redfn = getredfn(r) # get the pattern reduction function
            pats  = redfn(fltd) # reduce the filtered patterns
            pcs   = rpcs[r]     # get the pattern histogram and
            d = e = x = y = 0   # initialize the detection flags
            for p,c in pats:    # traverse the reduced patterns
                z = len(p)      # get pattern size and
                s = (z,int(c*5))# compute histogram bin
                if s in pcs: pcs[s] += 1 # count reduced patterns
                else:        pcs[s]  = 1 # by their signatures
                if   p >= asm:  # set detection flags, count patterns
                    if z > size:       sups[r] += 1; e = 1
                    else:                            d = 1
                elif p <  asm:         subs[r] += 1; x = 1
                elif len(p & asm) > 1: ovls[r] += 1; x = 1
                else:                  urls[r] += 1; x = 1
            dect[r] += d | e    # count runs with detections,
            xdct[r] += d        # runs with exact detections,
            rexs[r] += e        # runs with excess  neurons,
            rmis[r] += x        # runs with missing neurons,
            rops[r] += x | e    # runs with all other patterns
        srng += coin            # switch to the next random seed
    if verbose > 0:             # print termination log message
        print(strftime('%Y.%m.%d %H:%M:%S detect ')
                      +name+' '+str(runs)+' done.')
    for subdir in [join('raw','x'),'pu','pf','pr']:
        try: makedirs(join('data',subdir))
        except OSError: pass    # ensure that subdirectories exist
    n = 1.0/runs if runs > 0 else 1.0
    with open(join('data','raw','x','fn_'+name+'.txt'), 'w') as out:
        out.write('%g '  % (1.0 -draw*n))
        out.write('%g\n' % (1.0 -xraw*n))
    with open(join('data','pu','pu_'+name+'.txt'), 'w') as out:
        for x in sorted((s[0], s[1], upcs[s]*n) for s in upcs):
            out.write('%g %g %g\n' % x)
    with open(join('data','pf','pf_'+name+'.txt'), 'w') as out:
        for x in sorted((s[0], s[1], fpcs[s]*n) for s in fpcs):
            out.write('%g %g %g\n' % x)
    for r in rfnids:            # traverse the reduction function ids
        name = base+str(r)+ext  # build the file name and
        pcs  = rpcs[r]          # write the pattern counters
        try: makedirs(join('data','pr',str(r)))
        except OSError: pass    # ensure that directory exists
        outfn = join('data','pr',str(r),'pr_'+name+'.txt')
        with open(outfn, 'w') as out:
            for x in sorted((s[0], s[1], pcs[s]*n) for s in pcs):
                out.write('%g %g %g\n' % x)
        try: makedirs(join('data','raw',str(r)))
        except OSError: pass    # ensure that directory exists
        outfn = join('data','raw',str(r),'fn_'+name+'.txt')
        with open(outfn, 'w') as out:
            out.write('%g '  % (1.0 -dect[r]*n))
            out.write('%g '  % (1.0 -xdct[r]*n))
            out.write('%g '  % (rexs[r]*n)) # write the average
            out.write('%g '  % (rmis[r]*n)) # pattern counts
            out.write('%g '  % (rops[r]*n))
            out.write('%g '  % (sups[r]*n))
            out.write('%g '  % (subs[r]*n))
            out.write('%g '  % (ovls[r]*n))
            out.write('%g\n' % (urls[r]*n))
    return (upcs, fpcs, rpcs,   # return the pattern spectrums
            dect, xdct, rexs, rmis, rops, sups, subs, ovls, urls)

#-----------------------------------------------------------------------

def collect (ncnt=100, period=3, rate=20,          # neuron    params.
             sizes=[5], coins=[5], jitter=0.003,   # assembly  params.
             width=0.003, target='c', rfnids=[4]): # detection params.
    if jitter < 0: jitter = width
    base = 'n%d_p%g_r%g_j%g_w%g_%c' \
         % (ncnt,period,rate,jitter*10000,width*10000,target)
    fneg = dict()               # get the pattern count file name base
    fnxs = dict()               # and initialize run count dictionaries
    for s in [(z,c) for z in sizes for c in coins]:
        inpfn = join('data','raw','x','fn_'+base+('_%d_%d.txt' %s))
        with open(inpfn, 'r') as inp:
            fneg[s],fnxs[s] = inp.read()[:-1].split()
    for subdir in ['fn','fx','rx','rn','ro','sp','sb','ov','ur']:
        try: makedirs(join('data',subdir))
        except OSError: pass    # ensure that subdirectories exist
    with open(join('data','fn','fn_' +base +'.txt'), 'w') as out:
        for x in sorted([(k[0], k[1], fneg[k]) for k in fneg]):
            out.write('%g %g %s\n' % x)
    with open(join('data','fx','fx_' +base +'.txt'), 'w') as out:
        for x in sorted([(k[0], k[1], fnxs[k]) for k in fnxs]):
            out.write('%g %g %s\n' % x)
    for r in rfnids:            # traverse the reduction function ids
        name = base +str(r)     # get the pattern count file name
        fneg = dict(); fnxs = dict()
        rexs = dict(); rmis = dict(); rops = dict()
        sups = dict(); subs = dict(); ovls = dict(); urls = dict()
        for s in [(z,c) for z in sizes for c in coins]:
            outfn = join('data','raw',str(r),
                         'fn_' +name +('_%d_%d.txt' % s))
            with open(outfn, 'r') as inp:
                vals = inp.read()[:-1].split()
                fneg[s],fnxs[s]                 = vals[0:2]
                rexs[s],rmis[s],rops[s]         = vals[2:5]
                sups[s],subs[s],ovls[s],urls[s] = vals[5:9]
        with open(join('data','fn','fn_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], fneg[k]) for k in fneg]):
                out.write('%g %g %s\n' % x)
        with open(join('data','fx','fx_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], fnxs[k]) for k in fnxs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','rx','rx_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], rexs[k]) for k in rexs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','rn','rn_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], rmis[k]) for k in rmis]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ro','ro_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], rops[k]) for k in rops]):
                out.write('%g %g %s\n' % x)
        with open(join('data','sp','sp_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], sups[k]) for k in sups]):
                out.write('%g %g %s\n' % x)
        with open(join('data','sb','sb_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], subs[k]) for k in subs]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ov','ov_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], ovls[k]) for k in ovls]):
                out.write('%g %g %s\n' % x)
        with open(join('data','ur','ur_' +name +'.txt'), 'w') as out:
            for x in sorted([(k[0], k[1], urls[k]) for k in urls]):
                out.write('%g %g %s\n' % x)

#-----------------------------------------------------------------------

if __name__ == '__main__':
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( 'target', [  'c' ] ),
                ( {'jitter':0.0020, 'width':0.0040}, # 2.0 0.0020
                  {'jitter':0.0020, 'width':0.0038}, # 1.9
                  {'jitter':0.0020, 'width':0.0036}, # 1.8
                  {'jitter':0.0020, 'width':0.0034}, # 1.7
                  {'jitter':0.0020, 'width':0.0032}, # 1.6
                  {'jitter':0.0020, 'width':0.0030}, # 1.5
                  {'jitter':0.0020, 'width':0.0028}, # 1.4
                  {'jitter':0.0020, 'width':0.0026}, # 1.3
                  {'jitter':0.0020, 'width':0.0024}, # 1.2
                  {'jitter':0.0020, 'width':0.0022}, # 1.1
                  {'jitter':0.0020, 'width':0.0020}, # 1.0
                  {'jitter':0.0030, 'width':0.0060}, # 2.0 0.0030
                  {'jitter':0.0030, 'width':0.0057}, # 1.9
                  {'jitter':0.0030, 'width':0.0054}, # 1.8
                  {'jitter':0.0030, 'width':0.0051}, # 1.7
                  {'jitter':0.0030, 'width':0.0048}, # 1.6
                  {'jitter':0.0030, 'width':0.0045}, # 1.5
                  {'jitter':0.0030, 'width':0.0042}, # 1.4
                  {'jitter':0.0030, 'width':0.0039}, # 1.3
                  {'jitter':0.0030, 'width':0.0036}, # 1.2
                  {'jitter':0.0030, 'width':0.0033}, # 1.1
                  {'jitter':0.0030, 'width':0.0030}, # 1.0
                  {'jitter':0.0040, 'width':0.0080}, # 2.0 0.0040
                  {'jitter':0.0040, 'width':0.0076}, # 1.9
                  {'jitter':0.0040, 'width':0.0072}, # 1.8
                  {'jitter':0.0040, 'width':0.0068}, # 1.7
                  {'jitter':0.0040, 'width':0.0064}, # 1.6
                  {'jitter':0.0040, 'width':0.0060}, # 1.5
                  {'jitter':0.0040, 'width':0.0056}, # 1.4
                  {'jitter':0.0040, 'width':0.0052}, # 1.3
                  {'jitter':0.0040, 'width':0.0048}, # 1.2
                  {'jitter':0.0040, 'width':0.0044}, # 1.1
                  {'jitter':0.0040, 'width':0.0040}, # 1.0
                  {'jitter':0.0050, 'width':0.0100}, # 2.0 0.0050
                  {'jitter':0.0050, 'width':0.0095}, # 1.9
                  {'jitter':0.0050, 'width':0.0090}, # 1.8
                  {'jitter':0.0050, 'width':0.0085}, # 1.7
                  {'jitter':0.0050, 'width':0.0080}, # 1.6
                  {'jitter':0.0050, 'width':0.0075}, # 1.5
                  {'jitter':0.0050, 'width':0.0070}, # 1.4
                  {'jitter':0.0050, 'width':0.0065}, # 1.3
                  {'jitter':0.0050, 'width':0.0060}, # 1.2
                  {'jitter':0.0050, 'width':0.0055}, # 1.1
                  {'jitter':0.0050, 'width':0.0050}, # 1.0
                ),
                ( 'size',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'coin',   [ 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2 ] ),
                ( 'rfnids', [ range(getredfn(-1)) ] ),
                ( 'runs',   [ 1000 ] ) ] ]

    if len(argv) < 2 or argv[1] != 'collect':
        run(detect, grids, int(argv[1])-1 if len(argv) > 1 else -1)
    else:
        ranges = dict(project(grids, ['size','coin'])[0])
        grids  = project(grids, ['ncnt','period','rate','width',
                                 'jitter','target','rfnids'])
        for i in range(get_args(grids)):
            collect(sizes=ranges['size'], coins=ranges['coin'],
                    **get_args(grids, i))
